--- old/usr/src/lib/libzfs/common/libzfs_dataset.c	2013-04-25 12:27:55.780376180 +0000
+++ new/usr/src/lib/libzfs/common/libzfs_dataset.c	2013-04-25 12:27:55.786508480 +0000
@@ -4101,24 +4101,65 @@
 }
 
 int
+zfs_hold_add(zfs_handle_t *zhp, const char *snapname, const char *tag,
+    boolean_t enoent_ok, nvlist_t *holds)
+{
+	zfs_handle_t *szhp;
+	char name[ZFS_MAXNAMELEN];
+	char errbuf[1024];
+	int ret;
+
+	(void) snprintf(name, sizeof (name),
+	    "%s@%s", zhp->zfs_name, snapname);
+
+	szhp = make_dataset_handle(zhp->zfs_hdl, name);
+	if (szhp) {
+		fnvlist_add_string(holds, name, tag);
+		zfs_close(szhp);
+		return (0);
+	}
+
+	ret = ENOENT;
+	if (enoent_ok)
+		return (ret);
+
+	(void) snprintf(errbuf, sizeof (errbuf),
+	    dgettext(TEXT_DOMAIN, "cannot hold snapshot '%s@%s'"),
+	    zhp->zfs_name, snapname);
+	(void) zfs_standard_error(zhp->zfs_hdl, ret, errbuf);
+
+	return (ret);
+}
+
+int
 zfs_hold(zfs_handle_t *zhp, const char *snapname, const char *tag,
     boolean_t recursive, boolean_t enoent_ok, int cleanup_fd)
 {
 	int ret;
 	struct holdarg ha;
-	nvlist_t *errors;
-	libzfs_handle_t *hdl = zhp->zfs_hdl;
-	char errbuf[1024];
-	nvpair_t *elem;
 
 	ha.nvl = fnvlist_alloc();
 	ha.snapname = snapname;
 	ha.tag = tag;
 	ha.recursive = recursive;
 	(void) zfs_hold_one(zfs_handle_dup(zhp), &ha);
-	ret = lzc_hold(ha.nvl, cleanup_fd, &errors);
+	ret = zfs_hold_apply(zhp, enoent_ok, cleanup_fd, ha.nvl);
 	fnvlist_free(ha.nvl);
 
+	return (ret);
+}
+
+int
+zfs_hold_apply(zfs_handle_t *zhp, boolean_t enoent_ok, int cleanup_fd, nvlist_t *holds)
+{
+	int ret;
+	nvlist_t *errors;
+	libzfs_handle_t *hdl = zhp->zfs_hdl;
+	char errbuf[1024];
+	nvpair_t *elem;
+
+	ret = lzc_hold(holds, cleanup_fd, &errors);
+
 	if (ret == 0)
 		return (0);
 
@@ -4176,13 +4217,6 @@
 	return (ret);
 }
 
-struct releasearg {
-	nvlist_t *nvl;
-	const char *snapname;
-	const char *tag;
-	boolean_t recursive;
-};
-
 static int
 zfs_release_one(zfs_handle_t *zhp, void *arg)
 {
@@ -4199,6 +4233,7 @@
 		nvlist_t *holds = fnvlist_alloc();
 		fnvlist_add_boolean(holds, ha->tag);
 		fnvlist_add_nvlist(ha->nvl, name, holds);
+		fnvlist_free(holds);
 		zfs_close(szhp);
 	}
 
